// statement
// =========
import * as core from '../common/core';
import * as signal from '../common/signal';
import * as types from '../common/types';
import * as parser from '../parse/parser';
import * as operation from './operation';
import * as expression from './expression';
import * as assignment from './assignment';
import * as iolib from './iolib';
export function evaluate(line, state) {
    switch (line.type) {
        case types.EmptyStatement:
            return visitEmptyStatement(line, state);
        case types.EmptyStatement:
            return visitEmptyStatement(line, state);
        case types.AssignmentStatement:
            return visitAssignmentStatement(line, state);
        case types.ValueStatement:
            return visitValueStatement(line, state);
        case types.SyntaxError:
            return visitSyntaxError(line, state);
        case types.IfStatement:
            return visitIfStatement(line, state);
        case types.ThenStatement:
            return visitThenStatement(line, state);
        case types.ElseStatement:
            return visitElseStatement(line, state);
        case types.ForLoop:
            return visitForLoop(line, state);
        case types.WhileLoop:
            return visitWhileLoop(line, state);
        case types.RepeatLoop:
            return visitRepeatLoop(line, state);
        case types.EndStatement:
            return visitEndStatement(line, state);
        case types.PauseStatement:
            return visitPauseStatement(line, state);
        case types.LabelStatement:
            return visitLabelStatement(line, state);
        case types.GotoStatement:
            return visitGotoStatement(line, state);
        case types.IncrementSkip:
            return visitIncrementSkip(line, state);
        case types.DecrementSkip:
            return visitDecrementSkip(line, state);
        case types.MenuStatement:
            return visitMenuStatement(line, state);
        case types.ProgramStatement:
            return visitProgramStatement(line, state);
        case types.ReturnStatement:
            return visitReturnStatement(line, state);
        case types.StopStatement:
            return visitStopStatement(line, state);
        case types.DelVarStatement:
            return visitDelVarStatement(line, state);
        case types.GraphStyleStatement:
            return visitGraphStyleStatement(line, state);
        case types.OpenLibStatement:
            return visitOpenLibStatement(line, state);
        case types.ExecLibStatement:
            return visitExecLibStatement(line, state);
        case types.Display:
            return visitDisplay(line, state);
        case types.Input:
            return visitInput(line, state);
        case types.Prompt:
            return visitPrompt(line, state);
        case types.DispGraph:
            return visitDispGraph(line, state);
        case types.DispTable:
            return visitDispTable(line, state);
        case types.Output:
            return visitOutput(line, state);
        case types.ClrHome:
            return visitClrHome(line, state);
        case types.ClrTable:
            return visitClrTable(line, state);
        case types.GetCalc:
            return visitGetCalc(line, state);
        case types.Get:
            return visitGet(line, state);
        case types.Send:
            return visitSend(line, state);
        default:
            return core.exhaustiveMatchingGuard(line);
    }
}
// ----- Statements -----
function visitEmptyStatement(_line, _state) {
    // do nothing
    return undefined;
}
function visitAssignmentStatement(line, state) {
    var value = expression.evaluate(line.value, state.mem);
    assignment.evaluate(line.assignable, value, state.mem);
    operation.assignAns(state.mem, value);
    return undefined;
}
function visitValueStatement(line, state) {
    operation.assignAns(state.mem, expression.evaluate(line.value, state.mem));
    return undefined;
}
function visitSyntaxError(_line, _state) {
    throw core.SyntaxError;
}
// ----- CTL -----
function visitIfStatement(line, state) {
    if (line.value === null) {
        throw core.ArgumentError;
    }
    state.ifResult = operation.isTruthy(expression.evaluate(line.value, state.mem));
    return undefined;
}
function visitThenStatement(_line, _state) {
    throw core.SyntaxError;
}
function visitElseStatement(_line, state) {
    var previousBlockIndex = state.blockStack.pop();
    if (previousBlockIndex === undefined) {
        throw core.SyntaxError;
    }
    var previousBlockLine = state.lines[previousBlockIndex];
    if (previousBlockLine === undefined) {
        throw core.libError('Else blockstack led to missing line!');
    }
    if (previousBlockLine.type === types.ThenStatement) {
        state.blockStack.push(state.i);
        state.falsyStackHeight = state.blockStack.length;
    }
    else {
        throw core.SyntaxError;
    }
    return undefined;
}
function visitForLoop(line, state) {
    if (line.variable === null || line.start === null || line.end === null) {
        throw core.ArgumentError;
    }
    var value = expression.evaluate(line.start, state.mem);
    assignment.evaluate(line.variable, value, state.mem);
    state.blockStack.push(state.i);
    if (!operation.isTruthy(expression.evaluate(operation.binaryOperation(line.variable, '<=', line.end), state.mem))) {
        state.falsyStackHeight = state.blockStack.length;
    }
    return undefined;
}
function visitWhileLoop(line, state) {
    if (line.value === null) {
        throw core.ArgumentError;
    }
    state.blockStack.push(state.i);
    if (!operation.isTruthy(expression.evaluate(line.value, state.mem))) {
        state.falsyStackHeight = state.blockStack.length;
    }
    return undefined;
}
function visitRepeatLoop(line, state) {
    if (line.value === null) {
        throw core.ArgumentError;
    }
    state.blockStack.push(state.i);
    return undefined;
}
function visitEndStatement(_line, state) {
    var source = state.blockStack.pop();
    if (source === undefined) {
        throw core.SyntaxError;
    }
    var sourceLine = state.lines[source];
    if (sourceLine === undefined) {
        throw core.libError('End blockstack led to missing line!');
    }
    switch (sourceLine === null || sourceLine === void 0 ? void 0 : sourceLine.type) {
        case types.ForLoop:
            if (sourceLine.end === null) {
                throw core.libError('End blockstack led to invalid For!');
            }
            increment(state.mem, sourceLine.variable, sourceLine.step !== null ? sourceLine.step : core.ONE);
            if (operation.isTruthy(expression.evaluate(operation.binaryOperation(sourceLine.variable, '<=', sourceLine.end), state.mem))) {
                state.blockStack.push(source);
                state.i = source;
            }
            break;
        case types.WhileLoop:
        case types.RepeatLoop:
            if (sourceLine.value === null) {
                throw core.libError('End blockstack led to invalid While/Repeat!');
            }
            if (operation.isTruthy(expression.evaluate(sourceLine.value, state.mem))) {
                state.blockStack.push(source);
                state.i = source;
            }
            break;
        case types.ThenStatement:
        case types.ElseStatement:
            break;
        default:
            throw core.libError("impossibleEndFrom'".concat(sourceLine.type));
    }
    return undefined;
}
function visitPauseStatement(_line, _state) {
    throw core.UnimplementedError;
}
function visitLabelStatement(_line, _state) {
    // do nothing
    return undefined;
}
function visitGotoStatement(line, state) {
    state.searchLabel = line.location;
    state.i = -1;
    return undefined;
}
function visitIncrementSkip(line, state) {
    if (line.variable === null || line.end === null) {
        throw core.ArgumentError;
    }
    increment(state.mem, line.variable, core.ONE);
    state.incrementDecrementResult = operation.isTruthy(expression.evaluate(operation.binaryOperation(line.variable, '<=', line.end), state.mem));
    return undefined;
}
function visitDecrementSkip(line, state) {
    if (line.variable === null || line.end === null) {
        throw core.ArgumentError;
    }
    increment(state.mem, line.variable, core.MINUSONE);
    state.incrementDecrementResult = operation.isTruthy(expression.evaluate(operation.binaryOperation(line.variable, '>=', line.end), state.mem));
    return undefined;
}
function visitMenuStatement(line, state) {
    if (line.title === null || line.choices.length === 0 || line.choices.length > 7) {
        throw core.ArgumentError;
    }
    iolib.stdout(operation.valueToString(expression.evaluate(line.title, state.mem), true), state.io);
    line.choices.forEach(function (choice, idx) {
        iolib.stdout("".concat(idx + 1, ":").concat(operation.valueToString(choice.option)), state.io);
    });
    iolib.onStdin(function (input) {
        var digit = operation.parseDigit(input);
        if (digit === undefined) {
            return true;
        }
        var choice = line.choices[digit - 1];
        if (choice === undefined) {
            return true;
        }
        state.searchLabel = choice.location;
        state.i = 0;
        if (state.resume === undefined) {
            throw core.libError('resume not defined');
        }
        state.resume();
        return false;
    }, state.io);
    return signal.SUSPEND;
}
function visitProgramStatement(_line, _state) {
    throw core.UnimplementedError;
}
function visitReturnStatement(_line, _state) {
    // TODO interaction with subprograms
    return signal.DONE;
}
function visitStopStatement(_line, _state) {
    return signal.DONE;
}
function visitDelVarStatement(line, state) {
    if (line.variable === null) {
        throw core.ArgumentError;
    }
    operation.deleteVariable(state.mem, line.variable);
    return undefined;
}
function visitGraphStyleStatement(_line, _state) {
    throw core.UnimplementedError;
}
function visitOpenLibStatement(_line, _state) {
    throw core.UnimplementedError;
}
function visitExecLibStatement(_line, _state) {
    throw core.UnimplementedError;
}
// ----- I/O -----
function visitDisplay(line, state) {
    if (line.value !== null) {
        iolib.stdout(operation.valueToString(expression.evaluate(line.value, state.mem)), state.io);
    }
    return undefined;
}
function visitInput(line, state) {
    if (line.variable === null) {
        if (line.text !== null) {
            throw core.ArgumentError;
        }
        throw core.UnimplementedError;
    }
    var text = '?';
    if (line.text !== null) {
        text = operation.valueToString(expression.evaluate(line.text, state.mem), true);
    }
    return getInput(text, line.variable, state, true);
}
function visitPrompt(line, state) {
    if (line.variable === null) {
        throw core.ArgumentError;
    }
    var text = operation.variableToString(line.variable);
    return getInput("".concat(text, "=?"), line.variable, state, false);
}
function visitDispGraph(_line, _state) {
    throw core.UnimplementedError;
}
function visitDispTable(_line, _state) {
    throw core.UnimplementedError;
}
function visitOutput(line, state) {
    if (line.row === null || line.column === null || line.value === null) {
        throw core.ArgumentError;
    }
    var row = expression.evaluate(line.row, state.mem);
    var column = expression.evaluate(line.column, state.mem);
    if (row.type !== types.NUMBER || column.type !== types.NUMBER) {
        throw core.DataTypeError;
    }
    if (row.float < 1 || row.float > state.rows || column.float < 1 || column.float > state.columns) {
        throw core.DomainError;
    }
    // TODO: respect rows and columns
    iolib.stdout(operation.valueToString(expression.evaluate(line.value, state.mem)), state.io);
    return undefined;
}
function visitClrHome(_line, _state) {
    throw core.UnimplementedError;
}
function visitClrTable(_line, _state) {
    throw core.UnimplementedError;
}
function visitGetCalc(_line, _state) {
    throw core.UnimplementedError;
}
function visitGet(_line, _state) {
    throw core.UnimplementedError;
}
function visitSend(_line, _state) {
    throw core.UnimplementedError;
}
// ----- Helpers -----
function increment(mem, variable, step) {
    if (!operation.hasVariable(mem, variable)) {
        throw core.UndefinedError;
    }
    var value = expression.evaluate({
        type: types.BINARY,
        operator: '+',
        left: variable,
        right: step,
    }, mem);
    assignment.evaluate(variable, value, mem);
}
function getInput(text, variable, state, allowStringLiterals) {
    iolib.stdout(text, state.io, false);
    iolib.onStdin(function (input) {
        if (input === null || input === undefined || input === '') {
            return true;
        }
        if (state.resume === undefined) {
            throw core.libError('resume not defined');
        }
        state.resume(function () {
            iolib.stdout(input, state.io);
            var value;
            if (variable.type === types.STRINGVARIABLE && allowStringLiterals) {
                value = { type: types.STRING, chars: input };
            }
            else {
                value = expression.evaluate(parser.parseExpression(input), state.mem);
            }
            // special case where a prompt for a numerical variable is interpreted as a list variable
            if (variable.type === types.VARIABLE && value.type === types.LIST) {
                variable = { type: types.LISTVARIABLE, name: "List".concat(variable.name), custom: true };
            }
            assignment.evaluate(variable, value, state.mem);
        });
        return false;
    }, state.io);
    return signal.SUSPEND;
}
