// daemon
// ======
import * as signal from '../common/signal';
import * as loader from '../inject/loader';
var loopMessageName = 'daemon-loop';
var exceptionName = 'daemon-exception';
var minimumDelay = 0.001; // 1 microsecond
var tasks = new Map();
var exceptions = [];
var running = false;
var nextTaskId = 0;
var maxExceptions = 1000;
var looper;
function startLooper() {
    looper === null || looper === void 0 ? void 0 : looper.on(loopMessageName, handleMessage);
    looper === null || looper === void 0 ? void 0 : looper.on(exceptionName, handleException);
}
function stopLooper() {
    looper === null || looper === void 0 ? void 0 : looper.off(loopMessageName, handleMessage);
    looper === null || looper === void 0 ? void 0 : looper.off(exceptionName, handleException);
}
function updateLooper(value) {
    looper = value;
}
loader.subscribe(loader.LOOPER, updateLooper);
var eventFactory;
var eventTarget;
function updateEventFactory(value) {
    eventFactory = value;
    eventTarget = eventFactory.createEventTarget();
}
loader.subscribe(loader.EVENT, updateEventFactory);
var perf;
function updatePerf(value) {
    perf = value;
}
loader.subscribe(loader.PERF, updatePerf);
function fireEvent(name) {
    eventTarget === null || eventTarget === void 0 ? void 0 : eventTarget.dispatchEvent(name);
}
function startIfNeeded() {
    if (running === false) {
        running = true;
        fireEvent('start');
        startLooper();
        looper.post(loopMessageName);
    }
}
function createTask(func, delay, runOnce, options) {
    var taskId = nextTaskId++;
    var debug = options.debug;
    if (debug) {
        console.debug("[Task ".concat(taskId, "] Start"));
    }
    tasks.set(taskId, {
        func: func,
        delay: Math.max(delay, minimumDelay),
        lastRun: undefined,
        runOnce: runOnce,
        stopOnException: true,
        suspended: false,
        debug: debug,
    });
    startIfNeeded();
    return taskId;
}
function resumeTask(taskId) {
    var task = tasks.get(taskId);
    if (task === undefined) {
        throw new Error("Error resuming: task '".concat(taskId, "' not found"));
    }
    if (task.debug) {
        console.debug("[Task ".concat(taskId, "] Resume"));
    }
    task.suspended = false;
    startIfNeeded();
}
;
function suspendTask(taskId) {
    var task = tasks.get(taskId);
    if (task === undefined) {
        throw new Error("Error suspending: task '".concat(taskId, "' not found"));
    }
    if (task.debug) {
        console.debug("[Task ".concat(taskId, "] Suspend"));
    }
    task.suspended = true;
}
;
function deleteTask(taskId) {
    var task = tasks.get(taskId);
    if (task === undefined) {
        throw new Error("Error deleting: task '".concat(taskId, "' not found"));
    }
    if (task.debug) {
        console.debug("[Task ".concat(taskId, "] Stop"));
    }
    tasks.delete(taskId);
}
;
function handleMessage() {
    var time = perf.now();
    var runningTaskCount = 0;
    var suspendedTaskCount = 0;
    tasks.forEach(function (task, taskId) {
        var runs;
        if (task.suspended === true) {
            runs = 0;
            suspendedTaskCount += 1;
        }
        else {
            runningTaskCount += 1;
            if (task.lastRun === undefined || task.runOnce) {
                runs = 1;
            }
            else {
                var timeSinceRun = time - task.lastRun;
                runs = Math.floor(timeSinceRun / task.delay);
            }
        }
        if (runs > 0) {
            task.lastRun = time;
        }
        for (var i = 0; i < runs; i++) {
            var result = void 0;
            try {
                result = task.func();
            }
            catch (ex) {
                result = signal.FAULTED;
                if (exceptions.length < maxExceptions) {
                    exceptions.push(ex);
                    looper.post(exceptionName);
                }
            }
            if (result === signal.DONE ||
                task.runOnce ||
                (task.stopOnException && result === signal.FAULTED)) {
                deleteTask(taskId);
                runningTaskCount -= 1;
                break;
            }
            if (result === signal.SUSPEND) {
                suspendTask(taskId);
                runningTaskCount -= 1;
                suspendedTaskCount += 1;
                break;
            }
            if (result === signal.YIELD) {
                break;
            }
        }
    });
    if (runningTaskCount === 0) {
        running = false;
        stopLooper();
        if (suspendedTaskCount > 0) {
            fireEvent('suspend');
        }
        else {
            fireEvent('stop');
        }
    }
    else {
        looper.post(loopMessageName);
    }
}
;
function handleException() {
    if (exceptions.length > 0) {
        throw exceptions.pop();
    }
}
export function setTinyInterval(func, delay, options) {
    if (options === void 0) { options = { debug: false }; }
    return createTask(func, delay, false, options);
}
export function clearTinyInterval(tinyIntervalID) {
    deleteTask(tinyIntervalID);
}
export function resumeTinyInterval(tinyIntervalID) {
    resumeTask(tinyIntervalID);
}
export function setTinyTimeout(func, delay, options) {
    if (options === void 0) { options = { debug: false }; }
    return createTask(func, delay, true, options);
}
export function clearTinyTimeout(tinyTimeoutID) {
    deleteTask(tinyTimeoutID);
}
export function on(type, listener) {
    eventTarget.addEventListener(type, listener);
}
export function off(type, listener) {
    eventTarget.removeEventListener(type, listener);
}
